use std::{marker::PhantomData, ptr::NonNull};

use crate::trace::GCTrace;

pub struct GcRef<T: GCTrace>(pub(crate) NonNull<T>, PhantomData<T>);

impl<T: GCTrace> Clone for GcRef<T> {
	fn clone(&self) -> Self {
		Self(self.0, self.1)
	}
}

impl<T: GCTrace> GcRef<T> {
	pub(crate) unsafe fn new(ptr: NonNull<T>) -> Self {
		Self(ptr, PhantomData)
	}

	/// # Safety
	/// The caller needs to ensure that the underlying pointer hasn't been garbage collected.
	/// Since the drop order for garbage collected structs is undefined that means it is never
	/// safe to call this function in a [Drop::drop].
	///
	/// Do note that this doesn't mean that any particular instance of GcRef has to be marked
	/// as reachable during a call to [GCAllocator::gc][`crate::allocator::GCAllocator::gc`]
	/// but instead any GcRef instance referring to the same underlying pointer has to be reachable
	pub unsafe fn get(&self) -> &T {
		unsafe { self.0.as_ref() }
	}

	/// # Safety
	/// The caller needs to ensure that the underlying pointer hasn't been garbage collected.
	/// See [GcRef::get] for more details
	///
	/// The caller needs to ensure that this is the only instance of a pointer to the underlying value
	/// (in other words that [Clone] hasn't been called, or that all other clones of the pointer have
	/// already been dropped).
	///
	/// This function is hard(but not impossible) to use without causing UB. Unless you have a
	/// really special use case you might want to instead use one of various [cell][`std::cell`]
	/// types as the allocated type.
	pub unsafe fn get_mut(this: &mut Self) -> &mut T {
		this.0.as_mut()
	}
}
