use itertools::PeekingNext;

use super::operator::Operator;
use super::{
	token::{self, TokenType},
	ExpressionNode, Parser, Result,
};
use std::fmt::Debug;

pub struct BinaryExpr {
	pub left: Box<ExpressionNode>,
	pub operator: Operator,
	pub right: Box<ExpressionNode>,
}

impl BinaryExpr {
	pub fn new(left: Box<ExpressionNode>, operator: Operator, right: Box<ExpressionNode>) -> Self {
		Self {
			left,
			operator,
			right,
		}
	}
}

impl Debug for BinaryExpr {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		write!(f, "({} {:?} {:?})", self.operator, self.left, self.right)
	}
}

impl<'a, T: Iterator<Item = token::Token<'a>>> Parser<'a, T> {
	pub(super) fn equality(&mut self) -> Result<ExpressionNode> {
		let mut node = self.comparison()?;
		while let Some(o) = self
			.token_iter
			.peeking_next(|t| matches!(t.token_type, TokenType::EqualEqual | TokenType::BangEqual))
		{
			node = BinaryExpr::new(
				Box::new(node),
				o.token_type.try_into().unwrap(),
				Box::new(self.comparison()?),
			)
			.into();
		}
		Ok(node)
	}

	pub(super) fn comparison(&mut self) -> Result<ExpressionNode> {
		let mut node = self.term()?;

		while let Some(o) = self.token_iter.peeking_next(|t| {
			matches!(
				t.token_type,
				TokenType::Greater
					| TokenType::GreaterEqual
					| TokenType::Less | TokenType::LessEqual
			)
		}) {
			node = BinaryExpr::new(
				Box::new(node),
				o.token_type.try_into().unwrap(),
				Box::new(self.comparison()?),
			)
			.into();
		}
		Ok(node)
	}

	pub(super) fn term(&mut self) -> Result<ExpressionNode> {
		let mut node = self.factor()?;

		while let Some(o) = self
			.token_iter
			.peeking_next(|t| matches!(t.token_type, TokenType::Minus | TokenType::Plus))
		{
			node = BinaryExpr::new(
				Box::new(node),
				o.token_type.try_into().unwrap(),
				Box::new(self.comparison()?),
			)
			.into()
		}
		Ok(node)
	}

	pub(super) fn factor(&mut self) -> Result<ExpressionNode> {
		let mut node = self.unary()?;

		while let Some(o) = self
			.token_iter
			.peeking_next(|t| matches!(t.token_type, TokenType::Star | TokenType::Slash))
		{
			node = BinaryExpr::new(
				Box::new(node),
				o.token_type.try_into().unwrap(),
				Box::new(self.comparison()?),
			)
			.into();
		}
		Ok(node)
	}
}
